"""Commons."""


extensions = dict(JPEG="jpg", PNG="png", GTiff="tif", WEBP="webp")

drivers = dict(jpg="JPEG", png="PNG", tif="GTiff", webp="WEBP")

mimetype = dict(
    png="image/png",
    jpg="image/jpg",
    webp="image/webp",
    pbf="application/x-protobuf",
    mvt="application/x-protobuf",
)



"""Common response models."""

from starlette.responses import Response
from starlette.background import BackgroundTask


class TileResponse(Response):
    """Tiler's response."""

    def __init__(
        self,
        content: bytes,
        media_type: str,
        status_code: int = 200,
        headers: dict = {},
        background: BackgroundTask = None,
    ) -> None:
        """Init tiler response."""
        headers.update({"Content-Type": media_type})
        headers.update({"Cache-Control": "no-cache"})
        self.body = self.render(content)
        self.status_code = 200
        self.media_type = media_type
        self.background = background
        self.init_headers(headers)



"""rio_viz app."""

from typing import Any, Optional, Tuple, Union

import re
import urllib.parse
from functools import partial

from starlette.concurrency import run_in_threadpool

from rio_viz import version as rioviz_version

from rio_viz.raster import postprocess_tile
from rio_viz.models.mapbox import TileJSON
from rio_viz.ressources.enums import ImageType, VectorType
from rio_viz.ressources.common import drivers, mimetype
from rio_viz.ressources.responses import TileResponse
from rio_viz.templates.template import index_template_factory, simple_template_factory

from rio_tiler.utils import render
from rio_tiler.colormap import cmap
from rio_tiler.profiles import img_profiles

from starlette.requests import Request
from starlette.responses import Response, HTMLResponse
from starlette.templating import _TemplateResponse
from starlette.middleware.cors import CORSMiddleware
from starlette.middleware.gzip import GZipMiddleware

from fastapi import Depends, FastAPI, Query
import uvicorn

try:
    import rio_tiler_mvt  # noqa

    has_mvt = True
except ModuleNotFoundError:
    has_mvt = False


_postprocess_tile = partial(run_in_threadpool, postprocess_tile)
_render = partial(run_in_threadpool, render)


class viz(object):
    """Creates a very minimal slippy map tile server using fastAPI + Uvicorn."""

    def __init__(
        self,
        raster,
        token: str = None,
        port: int = 8080,
        host: str = "127.0.0.1",
        style: str = "satellite",
    ):
        """Initialize app."""
        self.app = FastAPI(
            title="titiler",
            description="A lightweight Cloud Optimized GeoTIFF tile server",
            version=rioviz_version,
        )
        self.app.add_middleware(
            CORSMiddleware,
            allow_origins=["*"],
            allow_credentials=True,
            allow_methods=["GET"],
            allow_headers=["*"],
        )
        self.app.add_middleware(GZipMiddleware, minimum_size=0)

        self.raster = raster
        self.port = port
        self.host = host
        self.style = style
        self.token = token

        @self.app.get(
            "/tiles/{z}/{x}/{y}.pbf",
            responses={
                200: {
                    "content": {"application/x-protobuf": {}},
                    "description": "Return a Mapbox Vector Tile.",
                }
            },
            description="Read COG and return a tile",
        )
        async def _mvt(
            z: int,
            x: int,
            y: int,
            tilesize: int = 128,
            feature_type: str = Query(
                None, title="Feature type", regex="^(point)|(polygon)$"
            ),
            resampling_method: str = Query("nearest", title="rasterio resampling"),
        ):
            """Handle /mvt requests."""
            content = await self.raster.read_tile_mvt(
                z,
                x,
                y,
                tilesize=tilesize,
                resampling_method=resampling_method,
                feature_type=feature_type,
            )
            return TileResponse(content, media_type=mimetype["pbf"])

        @self.app.get(
            r"/tiles/{z}/{x}/{y}",
            responses={
                200: {
                    "content": {"image/png": {}, "image/jpg": {}, "image/webp": {}},
                    "description": "Return an image.",
                }
            },
            response_class=TileResponse,
            description="Read COG and return a tile",
        )
        @self.app.get(
            r"/tiles/{z}/{x}/{y}.{ext}",
            responses={
                200: {
                    "content": {"image/png": {}, "image/jpg": {}, "image/webp": {}},
                    "description": "Return an image.",
                }
            },
            response_class=TileResponse,
            description="Read COG and return a tile",
        )
        async def _tile(
            z: int,
            x: int,
            y: int,
            scale: int = Query(2, gt=0, lt=4),
            ext: ImageType = None,
            indexes: Any = Query(None, description="Coma (',') delimited band indexes"),
            rescale: Any = Query(
                None, description="Coma (',') delimited Min,Max bounds"
            ),
            color_formula: str = Query(None, description="rio-color formula"),
            color_map: str = Query(None, description="rio-tiler color map names"),
            resampling_method: str = Query(
                "bilinear", description="rasterio resampling"
            ),
        ):
            """Handle /tiles requests."""
            if isinstance(indexes, str):
                indexes = tuple(int(s) for s in re.findall(r"\d+", indexes))

            tilesize = scale * 256
            tile, mask = self.raster.read_tile(
                z,
                x,
                y,
                tilesize=tilesize,
                indexes=indexes,
                resampling_method=resampling_method,
            )

            tile = await _postprocess_tile(
                tile, mask, rescale=rescale, color_formula=color_formula
            )

            if not ext:
                ext = ImageType.jpg if mask.all() else ImageType.png

            driver = drivers[ext]
            options = img_profiles.get(driver.lower(), {})
            options["colormap"] = (
                cmap.get(color_map) if color_map else self.raster.colormap
            )

            content = await _render(tile, mask, img_format=driver, **options)
            return TileResponse(content, media_type=mimetype[ext.value])

        @self.app.get(
            "/tilejson.json",
            response_model=TileJSON,
            responses={200: {"description": "Return a tilejson"}},
            response_model_exclude_none=True,
        )
        def _tilejson(
            request: Request,
            response: Response,
            tile_format: Optional[Union[ImageType, VectorType]] = None,
        ):
            """Handle /tilejson.json requests."""
            scheme = request.url.scheme
            host = request.headers["host"]

            kwargs = dict(request.query_params)
            kwargs.pop("tile_format", None)

            tile_url = f"{scheme}://{host}/tiles/{{z}}/{{x}}/{{y}}"
            if tile_format:
                tile_url += f".{tile_format}"

            qs = urllib.parse.urlencode(list(kwargs.items()))
            if qs:
                tile_url += f"?{qs}"

            return dict(
                bounds=self.raster.bounds,
                center=self.raster.center,
                minzoom=self.raster.minzoom,
                maxzoom=self.raster.maxzoom,
                name="rio-viz",
                tilejson="2.1.0",
                tiles=[tile_url],
            )

        @self.app.get(
            "/metadata",
            responses={200: {"description": "Return the metadata of the COG."}},
        )
        async def _metadata(
            response: Response,
            pmin: float = 2.0,
            pmax: float = 98.0,
            indexes: Any = Query(None, title="Coma (',') delimited band indexes"),
        ):
            """Handle /metadata requests."""
            if isinstance(indexes, str):
                indexes = tuple(int(s) for s in re.findall(r"\d+", indexes))

            return self.raster.metadata(pmin, pmax, indexes=indexes)

        @self.app.get(
            "/info",
            responses={200: {"description": "Return the metadata of the COG."}},
        )
        def _info(response: Response,):
            """Handle /info requests."""
            return self.raster.info()

        @self.app.get(
            "/point", responses={200: {"description": "Return a point value."}}
        )
        def _point(response: Response, coordinates: Any):
            """Handle /point requests."""
            if isinstance(coordinates, str):
                coordinates = list(map(float, coordinates.split(",")))

            return self.raster.point(coordinates)

        @self.app.get(
            "/index.html",
            responses={200: {"description": "Simple COG viewer."}},
            response_class=HTMLResponse,
        )
        def _viewer(template: _TemplateResponse = Depends(index_template_factory)):
            """Handle /index.html."""
            template.context.update(
                {
                    "mapbox_access_token": self.token,
                    "mapbox_style": self.style,
                    "allow_3d": has_mvt,
                }
            )
            return template

        @self.app.get(
            "/index_simple.html",
            responses={200: {"description": "Simple COG viewer."}},
            response_class=HTMLResponse,
        )
        def _simple_viewer(
            template: _TemplateResponse = Depends(simple_template_factory),
        ):
            """Handle /index_simple."""
            template.context.update(
                {"mapbox_access_token": self.token, "mapbox_style": self.style}
            )
            return template

    def get_endpoint_url(self) -> str:
        """Get endpoint url."""
        return f"http://{self.host}:{self.port}"

    def get_template_url(self) -> str:
        """Get simple app template url."""
        return f"http://{self.host}:{self.port}/index.html"

    def get_simple_template_url(self) -> str:
        """Get simple app template url."""
        return f"http://{self.host}:{self.port}/index_simple.html"

    def get_bounds(self) -> str:
        """Get RasterTiles bounds."""
        return self.raster.bounds

    def get_center(self) -> Tuple:
        """Get RasterTiles center."""
        return self.raster.center

    def start(self):
        """Start tile server."""
        uvicorn.run(app=self.app, host=self.host, port=self.port, log_level="info")



"""Titiler Enums."""

from enum import Enum


class ImageType(str, Enum):
    """Image Type Enums."""

    png = "png"
    npy = "npy"
    jpg = "jpg"


class VectorType(str, Enum):
    """Vector Type Enums."""

    pbf = "pbf"
    mvt = "mvt"



"""rio-viz: Template Factory."""

import os
from typing import Callable

from starlette.requests import Request
from starlette.templating import Jinja2Templates, _TemplateResponse

html_templates = Jinja2Templates(directory=os.path.dirname(__file__))


def create_simple_template_factory(
    tilejson: str = "_tilejson", info: str = "_info"
) -> Callable[[Request], _TemplateResponse]:
    """
    Create a dependency which may be injected into a FastAPI app.  The input parameters are used to look up the specific
    URLs for each route.
    """

    def _template(request: Request) -> _TemplateResponse:
        """Create a template of ``templates/simple.html`` from a request"""
        return html_templates.TemplateResponse(
            name="simple.html",
            context={
                "request": request,
                # https://github.com/encode/starlette/blob/master/docs/routing.md#reverse-url-lookups
                # `request.url_for` will reconstruct the full URL path based on
                # the `name` of the route which here is the function name.
                "tilejson_endpoint": request.url_for(tilejson),
                "info_endpoint": request.url_for(info),
                # Mapbox parameters default to those defined in the CLI
                # These are overridden in the app itself
                "mapbox_token": "",
                "mapbox_style": "basic",
            },
            media_type="text/html",
        )

    return _template


def create_index_template_factory(
    tilejson: str = "_tilejson", metadata: str = "_metadata", point: str = "_point",
) -> Callable[[Request], _TemplateResponse]:
    """
    Create a dependency which may be injected into a FastAPI app.  The input parameters are used to look up the specific
    URLs for each route.
    """

    def _template(request: Request) -> _TemplateResponse:
        """Create a template of ``templates/index.html`` from a request"""
        return html_templates.TemplateResponse(
            name="index.html",
            context={
                "request": request,
                # https://github.com/encode/starlette/blob/master/docs/routing.md#reverse-url-lookups
                # `request.url_for` will reconstruct the full URL path based on
                # the `name` of the route which here is the function name.
                "tilejson_endpoint": request.url_for(tilejson),
                "metadata_endpoint": request.url_for(metadata),
                "point_endpoint": request.url_for(point),
                # Mapbox parameters default to those defined in the CLI
                # These are overridden in the app itself
                "mapbox_token": "",
                "mapbox_style": "basic",
                # allow_3d will be overridden in the app itself
                "allow_3d": "False",
            },
            media_type="text/html",
        )

    return _template


simple_template_factory = create_simple_template_factory()
index_template_factory = create_index_template_factory()



"""rio_viz.raster: raster tiles object."""

from typing import Dict, BinaryIO, Tuple, Union, Sequence, Optional

import re
from pathlib import Path
from functools import partial
from concurrent import futures

import numpy

from rio_tiler.io import COGReader
from rio_tiler.io.cogeo import multi_metadata, multi_tile, multi_point

from rio_tiler.utils import linear_rescale, _chunks

from rio_color.operations import parse_operations
from rio_color.utils import scale_dtype, to_math_type

from starlette.concurrency import run_in_threadpool


def _get_info(src_path: str) -> Dict:
    with COGReader(src_path) as cog:
        info = cog.info.copy()
        info["dtype"] = cog.dataset.meta["dtype"]
        info["band_descriptions"] = [band for _, band in info["band_descriptions"]]
    return info


def postprocess_tile(
    tile: numpy.ndarray,
    mask: numpy.ndarray,
    rescale: str = None,
    color_formula: str = None,
) -> numpy.ndarray:
    """Post-process tile data."""
    if rescale:
        rescale_arr = list(map(float, rescale.split(",")))
        rescale_arr = list(_chunks(rescale_arr, 2))
        if len(rescale_arr) != tile.shape[0]:
            rescale_arr = ((rescale_arr[0]),) * tile.shape[0]

        for bdx in range(tile.shape[0]):
            tile[bdx] = numpy.where(
                mask,
                linear_rescale(
                    tile[bdx], in_range=rescale_arr[bdx], out_range=[0, 255]
                ),
                0,
            )
        tile = tile.astype(numpy.uint8)

    if color_formula:
        # make sure one last time we don't have
        # negative value before applying color formula
        tile[tile < 0] = 0
        for ops in parse_operations(color_formula):
            tile = scale_dtype(ops(to_math_type(tile)), numpy.uint8)

    return tile


class RasterTiles(object):
    """Raster tiles object."""

    def __init__(
        self,
        src_path: Sequence[str],
        nodata: Union[str, int, float] = None,
        minzoom: Optional[int] = None,
        maxzoom: Optional[int] = None,
    ):
        """Initialize RasterTiles object."""
        if isinstance(src_path, str):
            src_path = (src_path,)

        self.path = src_path
        self.nodata = nodata

        with futures.ThreadPoolExecutor() as executor:
            responses = list(executor.map(_get_info, self.path))

        self.bounds = responses[0]["bounds"]
        self.center = responses[0]["center"]
        self.minzoom = minzoom if minzoom is not None else responses[0]["minzoom"]
        self.maxzoom = maxzoom if maxzoom is not None else responses[0]["maxzoom"]
        self.data_type = responses[0]["dtype"]
        self.colormap = responses[0].get("colormap")

        if len(self.path) != 1:
            self.band_descriptions = [
                Path(p).stem
                if re.match(r"^band[0-9]+$", responses[idx]["band_descriptions"][0])
                else responses[idx]["band_descriptions"][0]
                for idx, p in enumerate(self.path)
            ]
        else:
            self.band_descriptions = responses[0]["band_descriptions"]

    def info(self) -> Dict:
        """Return general info about the images."""
        return dict(
            bounds=self.bounds,
            center=self.center,
            minzoom=self.minzoom,
            maxzoom=self.maxzoom,
            band_descriptions=[
                (ix + 1, bd) for ix, bd in enumerate(self.band_descriptions)
            ],
            dtype=self.data_type,
        )

    def read_tile(
        self,
        z: int,
        x: int,
        y: int,
        tilesize: int = 256,
        indexes: Sequence[int] = None,
        resampling_method: str = "bilinear",
    ) -> Tuple[numpy.ndarray, numpy.ndarray]:
        """Read raster tile data and mask."""
        if len(self.path) != 1 and indexes:
            path: Sequence[str] = [self.path[ii - 1] for ii in indexes]
            indexes = None
        else:
            path = self.path

        return multi_tile(
            path,
            x,
            y,
            z,
            tilesize=tilesize,
            indexes=indexes,
            nodata=self.nodata,
            resampling_method=resampling_method,
        )

    async def read_tile_mvt(
        self,
        z: int,
        x: int,
        y: int,
        tilesize: int = 128,
        resampling_method: str = "bilinear",
        feature_type: str = "point",
    ) -> BinaryIO:
        """Read raster tile data and encode to MVT."""
        from rio_tiler_mvt import mvt

        mvt_encoder = partial(run_in_threadpool, mvt.encoder)

        tile, mask = self.read_tile(
            z, x, y, tilesize=tilesize, resampling_method=resampling_method
        )
        return await mvt_encoder(
            tile, mask, self.band_descriptions, feature_type=feature_type
        )

    def point(self, coordinates: Tuple[float, float]) -> dict:
        """Read point value."""
        results = multi_point(self.path, *coordinates)
        if len(self.path) != 1:
            results = [r[0] for r in results]
        else:
            results = results[0]

        return {
            "coordinates": coordinates,
            "value": {b: r for b, r in zip(self.band_descriptions, results)},
        }

    def metadata(
        self, pmin: float = 2.0, pmax: float = 98.0, indexes: Sequence[int] = None,
    ) -> dict:
        """Get Raster metadata."""
        if indexes:
            band_descriptions = [(ix, self.band_descriptions[ix - 1]) for ix in indexes]
        else:
            band_descriptions = [
                (ix + 1, d) for ix, d in enumerate(self.band_descriptions)
            ]

        if len(self.path) != 1 and indexes:
            path: Sequence[str] = [self.path[ii - 1] for ii in indexes]
            indexes = None
        else:
            path = self.path

        hist_options = dict()
        if self.colormap:
            hist_options.update(
                dict(bins=[k for k, v in self.colormap.items() if v != (0, 0, 0, 255)])
            )

        results = multi_metadata(
            path,
            pmin,
            pmax,
            indexes=indexes,
            nodata=self.nodata,
            hist_options=hist_options,
        )

        info = {"bounds": self.bounds, "minzoom": self.minzoom, "maxzoom": self.maxzoom}
        info["band_descriptions"] = band_descriptions

        if len(self.path) != 1:
            info["statistics"] = {
                band_descriptions[ix][0]: r["statistics"][1]
                for ix, r in enumerate(results)
            }
        else:
            info["statistics"] = results[0]["statistics"]

        info["dtype"] = self.data_type

        return info



"""rio_viz.cli."""

import os
import tempfile
from contextlib import contextmanager, ExitStack

import numpy

import click
from rio_viz import app, raster

from rio_cogeo.cogeo import cog_validate, cog_translate
from rio_cogeo.profiles import cog_profiles


@contextmanager
def TemporaryRasterFile(dst_path, suffix=".tif"):
    """Create temporary file."""
    fileobj = tempfile.NamedTemporaryFile(
        dir=os.path.dirname(dst_path), suffix=suffix, delete=False
    )
    fileobj.close()
    try:
        yield fileobj
    finally:
        os.remove(fileobj.name)


class MbxTokenType(click.ParamType):
    """Mapbox token type."""

    name = "token"

    def convert(self, value, param, ctx):
        """Validate token."""
        try:
            if not value:
                return ""

            assert value.startswith("pk")
            return value

        except (AttributeError, AssertionError):
            raise click.ClickException(
                "Mapbox access token must be public (pk). "
                "Please sign up at https://www.mapbox.com/signup/ to get a public token. "
                "If you already have an account, you can retreive your "
                "token at https://www.mapbox.com/account/."
            )


class NodataParamType(click.ParamType):
    """Nodata inddex type."""

    name = "nodata"

    def convert(self, value, param, ctx):
        """Validate and parse band index."""
        try:
            if value.lower() == "nan":
                return numpy.nan
            elif value.lower() in ["nil", "none", "nada"]:
                return None
            else:
                return float(value)
        except (TypeError, ValueError):
            raise click.ClickException("{} is not a valid nodata value.".format(value))


@click.command()
@click.argument("src_paths", type=str, nargs=-1, required=True)
@click.option(
    "--nodata",
    type=NodataParamType(),
    metavar="NUMBER|nan",
    help="Set nodata masking values for input dataset.",
)
@click.option(
    "--minzoom", type=int, help="Overwrite minzoom",
)
@click.option(
    "--maxzoom", type=int, help="Overwrite maxzoom",
)
@click.option(
    "--style",
    type=click.Choice(["satellite", "basic"]),
    default="basic",
    help="Mapbox basemap",
)
@click.option("--port", type=int, default=8080, help="Webserver port (default: 8080)")
@click.option(
    "--host",
    type=str,
    default="127.0.0.1",
    help="Webserver host url (default: 127.0.0.1)",
)
@click.option(
    "--mapbox-token",
    type=MbxTokenType(),
    metavar="TOKEN",
    default=lambda: os.environ.get("MAPBOX_ACCESS_TOKEN", ""),
    help="Pass Mapbox token",
)
@click.option("--no-check", is_flag=True, help="Ignore COG validation")
@click.option("--simple", is_flag=True, default=False, help="Launch simple viewer")
def viz(
    src_paths,
    nodata,
    minzoom,
    maxzoom,
    style,
    port,
    host,
    mapbox_token,
    no_check,
    simple,
):
    """Rasterio Viz cli."""
    # Check if cog
    src_paths = list(src_paths)
    with ExitStack() as ctx:
        for ii, src_path in enumerate(src_paths):
            if not no_check and not cog_validate(src_path):
                # create tmp COG
                click.echo("create temporaty COG")
                tmp_path = ctx.enter_context(TemporaryRasterFile(src_path))

                output_profile = cog_profiles.get("deflate")
                output_profile.update(dict(blockxsize="256", blockysize="256"))
                config = dict(
                    GDAL_TIFF_INTERNAL_MASK=os.environ.get(
                        "GDAL_TIFF_INTERNAL_MASK", True
                    ),
                    GDAL_TIFF_OVR_BLOCKSIZE="128",
                )
                cog_translate(src_path, tmp_path.name, output_profile, config=config)
                src_paths[ii] = tmp_path.name

        src_dst = raster.RasterTiles(
            src_paths, nodata=nodata, minzoom=minzoom, maxzoom=maxzoom
        )
        application = app.viz(
            src_dst, token=mapbox_token, port=port, host=host, style=style
        )
        url = (
            application.get_simple_template_url()
            if simple
            else application.get_template_url()
        )
        click.echo(f"Viewer started at {url}", err=True)
        click.launch(url)
        application.start()
